/* Tunguska, ternary virtual machine
 *
 * Copyright (C) 2007,2008 Viktor Lofgren
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "tryte.h"
#include <stdlib.h>
#include <queue>
#include "display.h"
#include "interrupt.h"
#include "memory.h"
#ifndef machine_h
#define machine_h

class machine : public memory {
	public:
		machine();
		~machine();

		/* Instruction codes */
		enum {
		CLV=-40,     BRK, RTI, LDA, LDX, LDY, STA, STX, STY, TAX,
			TAY, TXA, TYA, TSX, TXS, PSH, PHP, PLL, PLP, AND,
			EOR, ORA, BIT, ADD, CMP, INC, INX, INY, DEC, DEX,
			DEY, ASL, LSR, ROL, ROR, JMP, JSR, RST, CLC, CLI, 
/*Note NOP=0 b.design*/	NOP, SEC, SEI, MLH, MLL, DIV, MOD, PLX, PLY, PHX,
		       	PHY, JCC, JCT, JCF, JEQ, JNE, JLT, JGT, JVC, JVS, 
			IVC, PRM, TSH, BUT, LAD, PGT, PGS, CAD, XAM, XAY,
			XAX, XYX,

			PAUSE  = 39, DEBUG = 40
		}; 

		/* Addressing modes */
		enum {  ABS=-4, IMMEDIATE=-3, 
			AX=-2, AY=-1, ACC=0, IMPLICIT=0,
	       	  	INDX=1, INDY=2, INDIRECT = 3, XY = 4 };

		/* Process flag shifts */
		enum { C = 0, G = 1, I = 2, B = 3, V = 4, PR = 5 };

		/* Process single instruction */
		void instruction();

		/* Quick OP-code-assembly. Takes addressing mode and operation and
		 * bakes it together into a complete opcode */
		static tryte qop(tryte mode, tryte op) { return (mode << 4) + op; }
	
		tryte A, X, Y, P, S, SP, PCL, PCH, CL;

		friend int main(int argc, char* argv[]);

		/* Assembler access */
		friend int addop(int opcode, int shift, machine& m, int pc);

		/* Peripherials */
		friend class display;
		friend class kboard;


		/* Very handy for debugging purposes */
		static char* opcode_to_string(int opcode);

		bool autorun;
		bool trace;

		enum { IRQ_KEYBOARD = 0,
		 	IRQ_CLOCK = 1,
	       		IRQ_KEYBREAK = 2,
			IRQ_ARITH = 3,	/* Arithmetic error, not implemented */
			IRQ_USER = 4,
			IRQ_DEBUG
		};


		/* Add an interrupt to the queue */
		void queue_interrupt(interrupt* i) { interrupt_queue->push(i); }
	protected:
		static char translation_table[80][10];
		static void init_translation_table();

		void run_interrupt();

		/* set processor flags */
		void pflag(const tryte& t);
		/* ... without touching carry */
		void pflagnoc(const tryte& t);

		/* debug opcode */
		void debug();
		void pause();

		/* All instruction functions */

		void nop(); 
		void lda(tryte *t); void ldx(tryte *t); void ldy(tryte *t); 
		void sta(tryte *t); void stx(tryte *t); void sty(tryte *t); 
		
		void tax(); void tay(); void txa(); 
		void tya(); void tsx(); void txs();

		void xam(tryte* t); void xax(); void xay(); void xyx();

		void psh(tryte* t); void php(); void phx(); void phy();
		void pll(tryte* t); void plp(); void plx(); void ply();

		void add(tryte* t); void cmp(tryte* t); 
		void inc(tryte* t); void inx(); void iny();
		void dec(tryte* t); void dex(); void dey();

		void _and(tryte *t);
		void eor(tryte *t);
		void ora(tryte *t);
		void bit(tryte* t);
		void tsh(tryte* t); 
		void prm(tryte* t);
		void but(tryte* t);

		void asl(tryte* t); void lsr(tryte* t);
		void rol(tryte* t); void ror(tryte* t);
		void jmp(const tryte* low, const tryte* high); 
		void jsr(const tryte* low, const tryte* high); 
		void rst(); 

		void jct(const tryte *t, const tryte* t2); 
		void jcf(const tryte *t, const tryte* t2); 
		void jcc(const tryte* t, const tryte* t2); 
		void jeq(const tryte* t, const tryte* t2); 
		void jgt(const tryte *t, const tryte* t2); 
		void jne(const tryte* t, const tryte* t2); 
		void jlt(const tryte* t, const tryte* t2); 
		void jvc(const tryte* t, const tryte* t2); 
		void jvs(const tryte* t, const tryte* t2);

		void clc(); void cli(); void clv(); void ivc();
		void sec(); void sed(); void sei();

		void lad(const tryte* t, const tryte* t2);
		void cad(const tryte* t, const tryte* t2);

		void pgt(const tryte* t);
		void pgs(const tryte* t);

		void mlh(tryte* t); void mll(tryte* t); void div(tryte* t); void mod(tryte* t);

		void rti();

		void soft_brk();

		/* Instructions end here */
		void brk(); 
		void branch(int p);

		std::queue<interrupt*>* interrupt_queue;
};

#endif
